import json
import time

class StateFile(object):

    ''' A simple stat storage object, a glorified dict '''

    def __init__(self, existing_dict={}):

        # The list of time.times()s indicating when we last rotated
        self.rotation_list = []
        # the time.time() we last ran
        self.last_run = None
        # a list of the last live edges
        self.last_live = []
        # A list of failures last run
        self.verification_failues = []
        # A list of mtimes for zonefiles
        self.zone_mtimes = {}

        # Restore any existing saved values - setting values above
        # this means that we can add new values to the state file
        # without any worries of them saving.
        for key, val in existing_dict.iteritems():
            setattr(self, key, val)

    def set_last_run(self):
        ''' update when we last ran. '''
        self.last_run = time.time()

    def to_json(self):
        ''' dump a representation of keys/vals for storage. '''
        return json.dumps(self.__dict__)

    def last_rotation(self):
        if self.rotation_list:
            return self.rotation_list[-1]
        else:
            return None

    def add_rotation(self, max_rotations, newtime=None):
        ''' Add a rotation, rotate out oldest value.

         newtime: optional epoch time to manually append instead of
        the current tiem.

        '''

        if len(self.rotation_list) == max_rotations:
            # We have too many historical rotations, remove one
            self.rotation_list = self.rotation_list[1:]

        if newtime:
            self.rotation_list.append(newtime)
        else:
            self.rotation_list.append(time.time())



import fcntl

def acquire_lock(lockfile):
    # lockfile should be an opened file in mode w

    try:
        fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        return False

    return True



'''
 Edgemanage constants. Not much to see here.
'''

# Default configuration path
CONFIG_PATH = "/etc/edgemanage/edgemanage.yaml"
# Timeout to fetch objects over HTTP
FETCH_TIMEOUT = 10
# Times to retry fetching an object if failed
FETCH_RETRY = 3

# Number of objects to store in fetch histories
FETCH_HISTORY = 2000

# Number of seconds that define a window within which to search for
# fetch entries. Used in a case where we want to check the last $N
# seconds worth of values fetched for a paritcular edge to give us a
# more accurate average than simply averaging the last FETCH_HISTORY
# entries.
DECISION_SLICE_WINDOW = 300

# Amount of time within which to check for a rotation in order to
# change Nagios state to WARNING.
NAGIOS_WARNING_TIME = 300

# Number of historical rotations to keep in the state file.
STATE_HISTORICAL_ROTATIONS = 100

# Upper domain to use for looking up IP addresses of edges while
# populating zone files
UPPER_DOMAIN = "deflect.ca"

# Valid modes that an edge can be set to
#  available - available for use normally
#  force - force an edge into service if it's eligible
#  blindforce - force an edge into service no matter what
#  unavailable - never use an edge
VALID_MODES = ["available", "force", "blindforce", "unavailable"]

# Valid states that an edge will be put into after a result of a
# decision being passed upon tests.
VALID_HEALTHS = ["pass", "fail", "pass_window", "pass_average"]



#!/usr/bin/env python

import unittest
from decisionmaker import DecisionMaker

import test_edgestate

from const import FETCH_TIMEOUT

TEST_EDGE = "testedge1"
GOOD_ENOUGH = 1.0

class DecisionMakerTest(unittest.TestCase):

    def _get_failing_edge_state(self):
        es = test_edgestate.EdgeStateTest._make_store()
        es.add_value(GOOD_ENOUGH * 2)
        return es

    def _get_passing_edge_state(self):
        es = test_edgestate.EdgeStateTest._make_store()
        es.add_value(GOOD_ENOUGH/10)
        return es

    def _get_error_edge_state(self):
        es = test_edgestate.EdgeStateTest._make_store()
        es.add_value(FETCH_TIMEOUT)
        return es

    def test_error_state(self):
        es = self._get_error_edge_state()
        dm = DecisionMaker()
        dm.add_edge_state(es)
        self.assertEqual(dm.check_threshold(GOOD_ENOUGH), {"fail": 1,
                                                           'pass_window': 0,
                                                           'pass_average': 0,
                                                           'pass': 0})

    def test_good_state(self):
        es = self._get_passing_edge_state()
        dm = DecisionMaker()
        dm.add_edge_state(es)
        self.assertEqual(dm.check_threshold(GOOD_ENOUGH), {"fail": 0,
                                                           'pass_window': 0,
                                                           'pass_average': 0,
                                                           'pass': 1})

    def test_fail_state(self):
        es = self._get_failing_edge_state()
        dm = DecisionMaker()
        dm.add_edge_state(es)
        self.assertEqual(dm.check_threshold(GOOD_ENOUGH), {"fail": 1,
                                                           'pass_window': 0,
                                                           'pass_average': 0,
                                                           'pass': 0})

    #def test_judgement(self):
    #    dm = DecisionMaker()
    #    passing_edge_state = _get_passing_edge_state()
    #    failing_edge_state = _get_failing_edge_state()

if __name__ == '__main__':
    unittest.main()



from const import DECISION_SLICE_WINDOW, VALID_HEALTHS, FETCH_TIMEOUT

import logging
import time

class DecisionMaker(object):

    def __init__(self):
        self.edge_states = {}
        # A results dict with edge as key, string as value, one of
        # VALID_HEALTHS
        self.current_judgement = {}

    def add_edge_state(self, edge_state):
        self.edge_states[edge_state.edgename] = edge_state
        self.current_judgement[edge_state.edgename] = None

    def get_judgement(self, edgename):
        return self.current_judgement[edgename]

    def check_threshold(self, good_enough):

        ''' Check fetch response times for being under the given
        threshold.

        '''

        # dict for stats to return
        results_dict = {}
        for statusname in VALID_HEALTHS:
            results_dict[statusname] = 0

        for edgename, edge_state in self.edge_states.iteritems():
            time_slice = edge_state[time.time() - DECISION_SLICE_WINDOW:time.time()]
            if time_slice:
                time_slice_avg = sum(time_slice)/len(time_slice)
                logging.debug("Analysing %s. Last val: %f, time slice: %f, average: %f",
                              edgename, edge_state.last_value(), time_slice_avg,
                              edge_state.current_average())
            else:
                time_slice_avg = None
                logging.debug("Analysing %s. Last val: %f, time slice: Not enough data, average: %f",
                              edgename, edge_state.last_value(), edge_state.current_average())

            if edge_state.last_value() < good_enough:
                logging.info("PASS: Last fetch for %s is under the threshold (%f < %f)", edgename,
                             edge_state.last_value(), good_enough)
                self.current_judgement[edgename] = "pass"
                results_dict["pass"] += 1
            elif edge_state.last_value() == FETCH_TIMEOUT:
                results_dict["fail"] += 1
                self.current_judgement[edgename] = "fail"
                logging.info(("FAIL: Fetch time for %s is equal to the FETCH_TIMEOUT of %d. "
                              "Automatic fail"),
                             edgename, FETCH_TIMEOUT)
            elif time_slice and time_slice_avg < good_enough:
                self.current_judgement[edgename] = "pass_window"
                results_dict["pass_window"] += 1
                logging.info("UNSURE: Last fetch for %s is NOT under the threshold but the average of the last %d items is (%f < %f)", edgename, len(time_slice), time_slice_avg, good_enough)
            elif edge_state.current_average() < good_enough:
                results_dict["pass_average"] += 1
                self.current_judgement[edgename] = "pass_average"
                logging.info("UNSURE: Last fetch for %s is NOT under the threshold but under the average (%f < %f)",
                             edgename, edge_state.current_average(), good_enough)
            else:
                results_dict["fail"] += 1
                self.current_judgement[edgename] = "fail"
                logging.info("FAIL: No check for %s has passed - last fetch time was %f",
                             edgename, edge_state.last_value())

        return results_dict



import os
import json
import time
import logging
import datetime
import copy

from const import FETCH_HISTORY, VALID_MODES, VALID_HEALTHS

ASSUMED_VALS={
    # A list of timestamps of when this edge has been in rotation
    "rotation_history": [],
    # A dict keyed by timestamps with values of floats containing
    # fetch times - limited to FETCH_HISTORY items
    "fetch_times": {},
    # A dict keyed by timestamps which keeps an average of fetch times
    # for FETCH_HISTORY days
    "historical_average": {},
    "state": "out",
    "mode": "available",
    "health": "pass",
    "state_entry_time": None,
    # A comment created by edge_conf when changing state
    "comment": "",
}

class EdgeState(object):

    def __init__(self, edgename, store_dir, nowrite=False):
        '''An object representing a simple set of time series data,
        backed by a local JSON file store. Also some state variables.

         Aka: Anything but RRD.
        '''

        self.edgename = edgename
        self.nowrite = nowrite
        self.statfile = os.path.join(store_dir, "%s.edgestore" % edgename)
        if os.path.isfile(self.statfile) and os.path.getsize(self.statfile) != 0:
            with open(self.statfile) as statfile_f:
                stat_info = json.load(statfile_f)
            for val_key, val_type in ASSUMED_VALS.iteritems():
                # Set self attributes for all dict vals in the stat
                # store.
                try:
                    setattr(self, val_key, stat_info[val_key])
                except KeyError as e:
                    # If the stat store lacks one of the keys in the
                    # dict, then initialise it with the default value
                    # from ASSUMED_VALS (usually just a type) - this
                    # lets us add new fields as we go along. Like a
                    # database migration for flat files :)
                    logging.error("Edgefile %s lacks %s, assuming %s", self.statfile,
                                  val_key, str(val_type))
                    setattr(self, val_key, copy.copy(val_type))
        else:
            # There is no stat file, just load empty assumed vals
            logging.warning("Initialising previously untracked edge %s", self.edgename)
            for val_key, val_type in ASSUMED_VALS.iteritems():
                setattr(self, val_key, copy.copy(val_type))

    def _dump(self):
        ''' Write out stat data to file '''
        output = {}

        if self.nowrite:
            logging.debug("Not writing %s because nowrite=True", self.statfile)
            return

        for val_key, val_type in ASSUMED_VALS.iteritems():
            output[val_key] = getattr(self, val_key)
        with open(self.statfile, "w") as statfile_f:
            json.dump(output, statfile_f, sort_keys=True, indent=4)

    def set_comment(self, comment):
        self.comment = comment
        self._dump()

    def unset_comment(self):
        self.comment = ""
        self._dump()

    def set_health(self, health):
        if health not in VALID_HEALTHS:
            raise ValueError("Health must be one f %s, not %s",
                             str(VALID_HEALTHS), health)
        else:
            if self.health != health:
                logging.debug("Setting health for edge %s to %s", self.edgename, health)
                self.health = health
                self._dump()

    def set_state(self, state):
        ''' Set the state of the edge - in or out '''
        if state not in ["in", "out"]:
            raise ValueError("State must be either in or out, not %s", state)
        else:
            if self.state != state:
                self.state = state
                self._dump()

    def set_mode(self, mode):
        ''' Set the mode of the edge '''
        if mode not in VALID_MODES:
            raise ValueError("Mode %s isn't in the set of valid modes (%s)",
                             mode, str(VALID_MODES))
        else:
            if self.mode != mode:
                self.state_entry_time = time.time()
                self.mode = mode
                self._dump()

    def current_average(self):
        ''' Return an average of the current live set of values '''
        return sum(self.fetch_times.values())/len(self.fetch_times)

    def __len__(self):
        ''' Return the number of values for fetch times we have '''
        return len(self.fetch_times)

    def __getitem__(self, index):
        '''Return the datetime at a given timestamp - or return a slice of
        dates between two timestamps
        '''
        if isinstance(index, slice):
            return_dict = {}
            for key, val in self.fetch_times.iteritems():
                if key >= index.start and key <= index.stop:
                    return_dict[key] = val

            return return_dict
        else:
            return self.fetch_times[index]

    def last_value(self):
        ''' Get the most recent value stored '''
        return self.fetch_times[max(self.fetch_times.keys())]

    def add_rotation(self):
        self.rotation_history.append(time.time())
        self._dump()

    def add_value(self, new_value, timestamp=None):
        '''Add a new value to the fetch times store and check if we
        need to make a historical average

        '''

        if timestamp:
            the_time = timestamp
        else:
            the_time = time.time()

        # HACK: for legacy reasons, we need to cast to string
        # here. It's stupid. Need to fix this in future versions with
        # migration path for old state files.
        self.fetch_times[str(the_time)] = new_value

        # prune our values if there's too many of them
        if len(self.fetch_times) > FETCH_HISTORY:
            min_value = sorted(self.fetch_times.keys())[0]
            logging.debug("Rotating out item with timestamp %s and value %f due to fetch cache being over %d items",
                          min_value, self.fetch_times[min_value], FETCH_HISTORY)
            del(self.fetch_times[min_value])

        the_time_datetime = datetime.datetime.utcfromtimestamp(the_time)
        if the_time_datetime.minute == 0:
            # prune our values if there's too many of them
            if len(self.historical_average) > FETCH_HISTORY:
                min_value = sorted(self.historical_average.keys())[0]
                del(self.historical_average[min_value])
            self.historical_average[the_time] = self.current_average()

        self._dump()
        return the_time



#!/usr/bin/env python

import unittest
import util
import tempfile
import subprocess

class UtilTest(unittest.TestCase):

    '''Tests for any utility functions'''

    def test_acquire_lock(self, do_subproc=False):
        '''Test for file locking, ensure we can and then can't lock a temporary file.
        '''
        temp_file = tempfile.mktemp()
        with open(temp_file, "w") as temp_file_f:

            self.assertTrue(util.acquire_lock(temp_file_f),
                            msg="Couldn't lock temporary file")

            # ayyyy lmao. Gotta have another process to get an flock
            # refused. I am a bad person and you should hate me.
            test_lock_fail = subprocess.Popen(
                ["/usr/bin/python", "-c", ("from util import acquire_lock; import sys; "
                "ret = acquire_lock(open('%s', 'w')); "
                 "sys.exit(0 if not ret else 1)") % temp_file])
            returncode = test_lock_fail.wait()
            self.assertEqual(returncode, 0,
                             msg="Could lock already locked temporary file")

if __name__ == '__main__':
    unittest.main()



# stdlib
import urlparse
import hashlib
import logging

# local
import const

# external
import requests

# Make requests stop logging so much. I love you but you need to shut
# up.
requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)

USER_AGENT="Edgemanage v2 (https://github.com/equalitie/edgemanage)"

class FetchFailed(Exception):
    def __init__(self, edgetest, fetch_host, fetch_object, reason):
        message = "Failed to fetch %s/%s from %s: %s" % (fetch_host, fetch_object,
                                                         edgetest.edgename, reason)
        super(FetchFailed, self).__init__(message)

        self.edgename = edgetest.edgename
        self.local_sum = edgetest.local_sum
        self.fetch_host = fetch_host
        self.fetch_object = fetch_object

class VerifyFailed(Exception):
    def __init__(self, edgetest, fetch_host, fetch_object, reason):
        message = "Failed to verify %s/%s from %s: %s != %s" % (
            fetch_host, fetch_object, edgetest.edgename,
            edgetest.local_sum, reason)
        super(VerifyFailed, self).__init__(message)

        self.edgename = edgetest.edgename
        self.local_sum = edgetest.local_sum
        self.fetch_host = fetch_host
        self.fetch_object = fetch_object


class EdgeTest(object):

    def __init__(self, edgename, local_sum):
        """
         edgename: FQDN string of the edge to be tested
         local_sum: the pre-computed known checksum of the object to be fetched
        """

        self.edgename = edgename
        self.local_sum = local_sum

    def fetch(self, fetch_host, fetch_object, proto="https", verify=False):
        """
         fetch_host: The Host header to use when fetching
         fetch_object: The path to the object to be fetched
        """
        try:
            response = requests.get(urlparse.urljoin(proto + "://" + self.edgename,
                                                     fetch_object),
                                    verify=verify,
                                    timeout=const.FETCH_TIMEOUT,
                                    headers = {"Host": fetch_host,
                                               "User-Agent": USER_AGENT})
        except requests.exceptions.Timeout as e:
            # Just assume it took the maximum amount of time
            return const.FETCH_TIMEOUT
        except requests.exceptions.ConnectionError as e:
            logging.error("Connection error when fetching from %s: %s", self.edgename, str(e))
            for i in range(const.FETCH_RETRY-1):
                logging.warning("Retrying connection to %s", self.edgename)
                try:
                    response = requests.get(urlparse.urljoin(proto + "://" + self.edgename,
                                                             fetch_object),
                                            verify=False,
                                            timeout=const.FETCH_TIMEOUT,
                                            headers = {"Host": fetch_host})
                    # Request was successful, stop retrying and
                    # continue
                    break
                except requests.exceptions.ConnectionError as e:
                    continue
            else:
                logging.error("Failed to connect to %s after retrying %d times",
                              self.edgename, const.FETCH_RETRY)
                # The loop finished without a break - we got more
                # connection errors. Let's bail and return the maximum
                # amount of time. for/else is weird.
                return const.FETCH_TIMEOUT

        if not response.ok:
            logging.error("Object fetch failed on %s", self.edgename)
            raise FetchFailed(self, fetch_host,
                              fetch_object, response.text)

        remote_hash = hashlib.md5(response.content).hexdigest()

        if remote_hash != self.local_sum:
            logging.error("Failed to verify hash on %s!!", self.edgename)
            raise VerifyFailed(self, fetch_host,
                               fetch_object, remote_hash)

        return response.elapsed.total_seconds()



#!/usr/bin/env python

import unittest
import tempfile
import time

import edgestate

TEST_EDGE = "testedge1"
TEST_FETCH_HISTORY = 4
# Patch edgestate's import of const
edgestate.FETCH_HISTORY = TEST_FETCH_HISTORY

class EdgeStateTest(unittest.TestCase):

    # TODO load test JSON file to ensure object creation

    @staticmethod
    def _make_store():
        store_dir = tempfile.mkdtemp()
        a = edgestate.EdgeState(TEST_EDGE, store_dir)
        return a

    def testStoreAverage(self):
        a = self._make_store()

        for i in range(4):
            a.add_value(2)

        average = a.current_average()
        self.assertEqual(average, 2)

    def testStoreRotation(self):
        a = self._make_store()

        for i in range(TEST_FETCH_HISTORY+1):
            a.add_value(2)
            time.sleep(0.01)

        self.assertEqual(len(a), TEST_FETCH_HISTORY)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import unittest
from edgelist import EdgeList

TEST_DOMAIN = "fakesite.deflect.ca"

class EdgeListTest(unittest.TestCase):

    def test_reject_bad_NS(self):
        a = EdgeList()
        self.assertRaises(Exception, a.generate_zone, ("equalit.ie", "test_data", {
            "ns_records": ["adns1.easydns.com",
                           "adns2.easydns.com"]
        }))

    def test_set_edge_live(self):
        a = EdgeList()
        a.add_edge("test1")
        a.set_edge_live("test1")
        self.assertTrue(a.edges["test1"]["live"])

    def test_len(self):
        a = EdgeList()
        a.add_edge("test1")
        a.add_edge("test2")
        a.add_edge("test3")
        self.assertEqual(len(a), 3)

    def test_liveness_operations(self):
        a = EdgeList()
        a.add_edge("test1", live=True)
        a.add_edge("test2")
        a.add_edge("test3")
        a.set_edge_live("test2")
        self.assertEqual(a.get_live_count(), 2)
        self.assertTrue(a.is_live("test1"))

    def test_state_operations(self):
        a = EdgeList()
        a.add_edge("test1")
        a.add_edge("test2", state="pass")
        a.add_edge("test3", state="pass")
        a.add_edge("test4", state="fail")
        # invalid state - edgelist don't care.
        a.add_edge("test5", state="satan")
        a.add_edge("test6", state="pass")

        self.assertEqual(a.get_state_stats(), {'fail': 1, None: 1, 'satan': 1, 'pass': 3})
        self.assertTrue(a.set_live_by_state("pass", 2))
        self.assertEqual(a.get_live_count(), 2)
        self.assertEqual(len(a.get_edges("pass")), 3)

    def test_generate_zonefile(self):
        a = EdgeList()
        a.add_edge("example.com")
        a.set_edge_live("example.com")
        new_zone = a.generate_zone("test.com", "test_data", {"ns_records": ["adns1.easydns.com."],
                                                             "soa_mailbox": "test.derp.com",
                                                             "soa_nameserver": "derpderpderp.com"
                                                         },
                                   serial_number=1234)
        with open("test_data/test.com.output") as known_zone_f:
            known_zone = known_zone_f.read()
        self.assertEqual(known_zone, new_zone)

if __name__ == '__main__':
    unittest.main()



import os
from setuptools import setup

setup(
    name = "edgemanage",
    version = "2.0.2",
    author = "Hugh Nowlan",
    author_email = "nosmo@nosmo.me",
    description = "HTTP availability management tool",
    license = "Hacktivismo Enhanced-Source Software License Agreement",
    keywords = "edgemanage deflect DNS",
    url = "http://github.com/equalitie/edgemanage",
    packages=['edgemanage'],
    package_data={'edgemanage': ['templates/*.j2']},
    install_requires=[
        "Jinja2",
        "setproctitle",
        "pyyaml",
        "futures",
        "requests"
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Topic :: Internet :: Name Service (DNS)",
        "Topic :: Utilities",
        ],
    scripts = ["edge_manage", "edge_query", "edge_conf"],
    )



import const
import util
from edgetest import *
from edgelist import EdgeList
from edgestate import EdgeState
from decisionmaker import DecisionMaker
from statefile import StateFile
from edgemanage import EdgeManage



#!/usr/bin/env python

from edgetest import EdgeTest, VerifyFailed
from edgestate import EdgeState
from decisionmaker import DecisionMaker
from edgelist import EdgeList

from concurrent.futures import ProcessPoolExecutor, as_completed
import glob
import hashlib
import logging
import os


def future_fetch(edgetest, testobject_host, testobject_path,
                 testobject_proto, testobject_verify):
    """Helper function to give us a return value that plays nice with as_completed"""

    try:
        fetch_result = edgetest.fetch(testobject_host, testobject_path,
                                      testobject_proto, testobject_verify)
    except VerifyFailed:
        fetch_result = False
    return {edgetest.edgename: fetch_result}


class EdgeManage(object):

    def _init_objects(self):
        # List of edges that will be made live
        self.edgelist_obj = EdgeList()
        # Object we will use to make a decision about edge liveness based
        # on the stat stores
        self.decision = DecisionMaker()

        self.edge_states = {}

        self.testobject_hash = self.get_testobject_hash()
        self.current_mtimes = self.zone_mtime_setup()

    def get_testobject_hash(self):
        # Hash the local copy of the object to be requested from the edges
        with open(self.config["testobject"]["local"]) as test_local_f:
            testobject_hash = hashlib.md5(test_local_f.read()).hexdigest()
            logging.info("Hash of local object %s is %s",
                         self.config["testobject"]["local"], testobject_hash)

        return testobject_hash

    def __init__(self, dnet, config, state, dry_run=False):
        '''
         Upper-level edgemanage object that is used to create
        lower-level edgemanage objects and accomplish the overall task
        of edge testing, rotation and zone file writing.

        '''

        self.dnet = dnet
        self.dry_run = dry_run
        self.config = config
        self.state_obj = state

        self._init_objects()

    def zone_mtime_setup(self):
        # Get a complete list of zone names
        current_mtimes = {}
        for zonefile in glob.glob("%s/%s/*.zone" % (self.config["zonetemplate_dir"], self.dnet)):
            zone_name = zonefile.split(".zone")[0].split("/")[-1]
            # And while we're here, let's get their mtimes
            current_mtime = int(os.stat(zonefile).st_mtime)
            current_mtimes[zone_name] = current_mtime
        return current_mtimes

    def add_edge_state(self, edge, edge_healthdata_path, nowrite=False):
        edge_state = EdgeState(edge, edge_healthdata_path, nowrite=nowrite)
        self.edge_states[edge] = edge_state

    def do_edge_tests(self):

        test_dict = self.config["testobject"]
        test_host = test_dict["host"]
        test_path = test_dict["uri"]
        test_proto = test_dict["proto"]
        test_verify = test_dict["verify"]

        edgescore_futures = []
        with ProcessPoolExecutor() as executor:
            for edgename in self.edge_states:
                edge_t = EdgeTest(edgename, self.testobject_hash)
                edgescore_futures.append(executor.submit(future_fetch,
                                                         edge_t, test_host,
                                                         test_path,
                                                         test_proto,
                                                         test_verify))

        verification_failues = []

        for f in as_completed(edgescore_futures):
            try:
                result = f.result()
            except Exception as e:
                # Do some shit here
                raise
            edge, value = result.items()[0]

            if value == False:
                verification_failues.append(edge)

            self.edge_states[edge].add_value(value)
            logging.info("Fetch time for %s: %f avg: %f",
                         edge, value,
                         self.edge_states[edge].current_average())

            # Skip edges that we have forced out of commission
            if self.edge_states[edge].mode == "unavailable":
                logging.debug("Skipping edge %s as its status has been set to unavailable", edge)
            else:
                # otherwise add it to the decision maker
                self.decision.add_edge_state(self.edge_states[edge])

        return verification_failues

    def check_last_live(self):

        # A list of edges that were in use last time that are still
        # healthy now.
        still_healthy = []

        if self.state_obj.last_live:
            logging.debug("Live edge list from previous run is %s",
                          self.state_obj.last_live)

        # Make sure that any edges that were in rotation are still
        # in a passing state. Discard any that are failing checks.
        for oldlive_edge in self.state_obj.last_live:
            if oldlive_edge in self.decision.current_judgement and \
               self.decision.get_judgement(oldlive_edge) == "pass":
                still_healthy.append(oldlive_edge)
            elif oldlive_edge not in self.decision.current_judgement:
                logging.warning(("Discarding previously live edge %s "
                                 "because it is no longer being checked",),
                                oldlive_edge)
            else:
                logging.debug(
                    "Discarding previously live edge %s because it is in state %s",
                    oldlive_edge,
                    self.decision.current_judgement[oldlive_edge])

        return list(set(still_healthy))

    def make_edges_live(self, force_update):

        # Returns true if any changes were made.

        good_enough = self.config["goodenough"]
        required_edge_count = self.config["edge_count"]
        if self.dnet in self.config["dnet_edge_count"]:
            required_edge_count = self.config["dnet_edge_count"][self.dnet]

        # Has the edgelist changed since last iteration?
        edgelist_changed = None
        # Have ANY changes happened since last iteration? Including zone
        # updates.
        any_changes = False

        threshold_stats = self.decision.check_threshold(good_enough)

        # Get the list of previously healthy edges
        still_healthy_from_last_run = self.check_last_live()

        for edgename, edge_state in self.edge_states.iteritems():
            if edge_state.mode == "force":
                if self.decision.get_judgement(edgename) == "pass":
                    logging.debug(
                        "Making host %s live because it is in mode force and it is in state pass",
                        edgename)

                    # Don't set edgelist_changed to True if we're
                    # already heathy and live
                    if not edgename in still_healthy_from_last_run:
                        self.edgelist_obj.add_edge(edgename, state="pass", live=True)
                        edgelist_changed = True

            elif edge_state.mode == "blindforce":
                logging.debug("Making host %s live because it is in mode blindforce.",
                              edgename)
                self.edgelist_obj.add_edge(edgename, state="pass", live=True)

                # Don't update the edgelist if we're still in the last
                # live list. We don't care if we're healthy.
                if edgename not in self.state_obj.last_live:
                    edgelist_changed = True

        logging.debug("Stats of threshold check are %s", str(threshold_stats))

        # If everything is still healthy from the last run then use those.
        if still_healthy_from_last_run:
            logging.info("Got list of previously in use edges that are in a passing state: %s",
                         still_healthy_from_last_run)
            if edgelist_changed == None:
                # This check is to ensure that a previously-passing
                # list doesn't ignore forced or blindforced edges.
                edgelist_changed = False

        for still_healthy in still_healthy_from_last_run:
            if len(self.edgelist_obj) < required_edge_count:
                self.edgelist_obj.add_edge(still_healthy, state="pass", live=True)

        if len(still_healthy_from_last_run) == required_edge_count:
            logging.info(
                "Old edge list is still healthy - not making any changes"
            )
        else:
            logging.debug(("Didn't have enough healthy edges from last run to meet "
                           "edge count - trying to add more edges"))
            edgelist_changed = True

            for decision_edge, edge_state in self.decision.current_judgement.iteritems():
                if decision_edge not in self.edgelist_obj.edges:
                    self.edgelist_obj.add_edge(decision_edge, state=edge_state)
            logging.debug("List of previously passing edges is currently %s",
                          self.edgelist_obj.get_live_edges())

            # Attempt to meet demand, first with passing, then with
            # window, then with average passing
            for desired_state in ["pass", "pass_window", "pass_average"]:
                filled_by_current_state = self.edgelist_obj.set_live_by_state(desired_state,
                                                                              required_edge_count)
                if filled_by_current_state:
                    logging.info("Filled requirement for %d edges with edges in state %s",
                                 required_edge_count, desired_state)
                    break
            else:
                # Entering an "else" in the context of a "for" loop means
                # "we didn't break". It's horrible but it's exactly what
                # we need here.
                logging.error("Tried to add edges from all acceptable states but failed")
                # TODO randomly try to add edges in a panic

        if self.edgelist_obj.get_live_count() == required_edge_count:
            logging.info("Successfully established %d edges: %s",
                         self.edgelist_obj.get_live_count(),
                         self.edgelist_obj.get_live_edges())

            # Iterate over every *zone file in the zonetemplate dir and write out files.
            for zone_name in self.current_mtimes:

                # Unless an update is forced:
                # * Skip files that haven't been changed
                # * Write out zone files we haven't seen before
                # * don't write out updated zone files when we aren't changing edge list
                old_mtime = self.state_obj.zone_mtimes.get(zone_name)
                if not force_update and not edgelist_changed and old_mtime and old_mtime == self.current_mtimes[zone_name]:
                    logging.info("Not writing zonefile for %s because there are no changes pending",
                                 zone_name)
                    continue
                else:
                    any_changes = True

                complete_zone_str = self.edgelist_obj.generate_zone(
                    zone_name, os.path.join(self.config["zonetemplate_dir"], self.dnet),
                    self.config["dns"]
                )

                complete_zone_path = os.path.join(self.config["named_dir"],
                                                  "%s.zone" % zone_name)
                #TODO add rotation of old files
                if not self.dry_run:
                    with open(complete_zone_path, "w") as complete_zone_f:
                        logging.debug("Writing completed zone file for %s to %s",
                                      zone_name, complete_zone_path)
                        complete_zone_f.write(complete_zone_str)
                else:
                    logging.debug(("In dry run so not writing file %s for zone %s. "
                                   "It would have contained:\n%s"),
                                  complete_zone_path, zone_name, complete_zone_str)


        else:
            logging.error("Couldn't establish full edge list! Only have %d edges (%s), need %d",
                          self.edgelist_obj.get_live_count(),
                          self.edgelist_obj.get_live_edges(),
                          required_edge_count)

        # We've got our edges, one way or another - let's set their states
        # Note in the statefile that this edge has been put into rotation
        for edge in self.edge_states:

            if self.edgelist_obj.is_live(edge):
                # Note in the statefile that this edge has been put into rotation
                logging.debug("Setting edge %s to state in", edge)
                self.edge_states[edge].add_rotation()
                self.edge_states[edge].set_state("in")
            else:
                logging.debug("Setting edge %s to state out", edge)
                self.edge_states[edge].set_state("out")

            if self.edge_states[edge].mode != "unavailable":
                current_health = self.decision.get_judgement(edge)
                self.edge_states[edge].set_health(current_health)

        self.state_obj.zone_mtimes = self.current_mtimes

        return any_changes or edgelist_changed



from const import DECISION_SLICE_WINDOW

import logging
import time
from collections import Counter
import random
import socket
import time
import os

from jinja2 import Environment, PackageLoader, FileSystemLoader

try:
    env = Environment(loader=PackageLoader('edgemanage','templates'))
except ImportError as e:
    # we're not installed as a module
    if os.path.exists("edgemanage/templates") and os.path.isdir("edgemanage/templates"):
        env = Environment(loader=FileSystemLoader("edgemanage/templates"))
    elif os.path.exists("templates") and os.path.isdir("templates"):
        env = Environment(loader=FileSystemLoader("templates"))
    else:
        raise

class EdgeList(object):
    """ A class that represents a list of edges """

    def __init__(self):
        # A dictionary indicating whether an edge is live or not, and
        # what state it's in
        self.edges = {}

    def add_edge(self, edgename, state=None, live=False):
        self.edges[edgename] = {
            "live": live,
            "state": state
        }

    def __len__(self):
        return len(self.edges.keys())

    def get_state_stats(self):
        counter = Counter([ edge["state"] for edge in self.edges.values() ])
        return dict(counter)

    def get_live_count(self):
        return len(self.get_live_edges())

    def get_live_edges(self):
        return self.get_edges_by_liveness(True)

    def get_unlive_edges(self):
        return self.get_edges_by_liveness(False)

    def is_live(self, edgename):
        if edgename in self.edges:
            return self.edges[edgename]["live"]
        else:
            return None

    def get_edges_by_liveness(self, islive):
        edge_list = []
        for edge, state_dict in self.edges.iteritems():

            if state_dict["live"] == islive:
                edge_list.append(edge)
        return sorted(edge_list)

    def set_edge_live(self, edgename):
        self.edges[edgename]["live"] = True

    def set_live_by_state(self, state, desired_count):

        # return value that indicates whether we satisfied the
        # condition of setting $desired_count hosts of $state to live
        met_demand = False
        edge_list = self.get_edges(state)
        random.shuffle(edge_list)
        for edge in edge_list:
            if self.get_live_count() == desired_count:
                logging.debug("Edgelist got enough (%d) edges in state %s",
                              desired_count, state)
                met_demand = True
                break
            self.set_edge_live(edge)

        return met_demand

    def get_edges(self, state=None):
        ''' return a list of edges, with the option to filter by state '''
        if state:
            return [ edge for edge in self.edges if self.edges[edge]['state'] == state ]
        else:
            return self.edges.keys()

    def get_random_edge(self, state=None):
        selected_edges = self.get_edges(state)
        random.shuffle(selected_edges)
        if state:
            return selected_edges[0]

    def generate_zone(self, domain, zonefile_dir, dns_config,
                      serial_number=None):
        logging.debug("Started generating zone for %s", domain)

        if not all([ i.endswith(".") for i in dns_config["ns_records"] ]):
            raise Exception(("Nameserver list is incorrectly formatted. Every"
                             " entry should end with a full stop"))

        rotate_zones = []
        if "rotate_zones" in dns_config:
            rotate_zones = dns_config["rotate_zones"]

        live_edge_ips = []
        for live_edge in self.get_live_edges():
            try:
                edge_ip = socket.gethostbyname(live_edge)
                live_edge_ips.append(edge_ip)
            except socket.gaierror as e:
                try:
                    # Retry resolution failures
                    edge_ip = socket.gethostbyname(live_edge)
                    live_edge_ips.append(edge_ip)
                except socket.gaierror as e:
                    logging.error(("Failed to resolve IP address for %s! Correct"
                                   " hostname or remove this IP address from rotation."),
                                  live_edge)
                    continue

        logging.debug("Writing zone file for %s, live edge list is %s",
                      domain, self.get_live_edges())

        with open(os.path.join(zonefile_dir, "%s.zone" % domain)) as zonefile_f:
            zonefile = zonefile_f.read()

        if not serial_number:
            serial_number = int(time.time())
        template = env.get_template('zonetemplate.j2')
        return template.render(
            zonefile=zonefile,
            domain=domain,
            live_edge_ips=live_edge_ips,
            rotate_zones=rotate_zones,
            serial_number=serial_number,
            soa_mailbox=dns_config["soa_mailbox"],
            soa_nameserver=dns_config["soa_nameserver"],
            nameservers=dns_config["ns_records"]
        )
